import HsUnix
import IO
import Monad
import List
import CDs

d_direct = argdesc [ desc_direct
                   , desc_any_times
                   ]

d_ersetzen = argdesc [ desc_at_most_once
                     , desc_long "ersetzen"
                     , desc_short 'e'
                     , desc_description "Substrings im Linkziel ersetzen"
                     ]

d_herholen = argdesc [ desc_at_most_once
                     , desc_long "herholen"
                     , desc_short 'h'
                     , desc_description "Symlink durch referierte Datei ersetzen"
                     ]

d_tief_loeschen = argdesc [ desc_at_most_once
                          , desc_long "tief-lschen"
                          , desc_short 't'
                          , desc_description "Symlink und referierte Datei lschen"
                          ]

d_tief_umbenennen = argdesc [ desc_at_most_once
                            , desc_long "tief-umbenennen"
                            , desc_short 'u'
                            , desc_description "Symlink und referierte Datei umbenennen"
                            ]

descs = [ d_direct, d_ersetzen, d_herholen, d_tief_loeschen, d_tief_umbenennen ]

data Funktion = Herholen | Ersetzen | Tief_loeschen | Tief_umbenennen

data Weiter = Weiter
instance Typeable Weiter where typeOf _ = mkAppTy tyCon_weiter []
tyCon_weiter = mkTyCon "Main.Weiter"


main = mainwrapper main'

main' = do
    (f,args) <- kzp
    case f of
       Herholen -> alle herholen args
       Ersetzen ->
          case args of
             (von:zu:symlinks) -> alle (ersetzen von zu) symlinks
             _                 -> hPutStrLn stderr "umlinken --ersetzen <von> <zu> <Symlinks>\n"
       Tief_loeschen -> alle tief_loeschen args
       Tief_umbenennen -> 
          case args of
             [alt,neu] -> tief_umbenennen alt neu
             _         -> hPutStrLn stderr "umlinken --tief-umbenennen <Symlink> <neuer Name>\n"

    where
       alle f args =
          mapM_ (\symlink -> f symlink
                               `catchDyn` weiter
                               `catchDyn` errnomeld symlink)
                args

       weiter :: Weiter -> IO ()
       weiter Weiter = return ()

       errnomeld :: String -> Errno -> IO ()
       errnomeld symlink errno = do
          str <- strerror errno
          hPutStrLn stderr $ symlink ++ ": " ++ str


herholen symlink = do
   pruefe_linkziel symlink
   pruefe_offen symlink         -- ob in Behlter wird nicht gebraucht
   linkziel <- readlink symlink
   rm symlink
   mv linkziel symlink

ersetzen von zu symlink = do
   pruefe_symlink symlink
   linkziel <- readlink symlink
   let linkziel' = ers von zu linkziel
   rm symlink
   HsUnix.symlink linkziel' symlink
   where
      -- Alle Vorkommen eines Substrings ersetzen.
      ers von zu [] = []
      ers von zu txt =
          if (isPrefixOf von txt)
              then let len  = length von
                       rest = drop len txt
                   in  zu ++ ers von zu rest
              else (head txt : ers von zu (tail txt))

tief_loeschen symlink = do
   pruefe_linkziel symlink
   linkziel <- readlink symlink
   rm linkziel
   rm symlink

tief_umbenennen symlink neu = do
   pruefe_linkziel symlink
   linkziel <- readlink symlink

   let (zielpfad,_)    = split_path linkziel
       (symlinkpfad,_) = split_path symlink
       neuziel         = zielpfad ++ "/" ++ neu
       neusymlink      = symlinkpfad ++ "/" ++ neu

   when (fst (split_path neu) /= ".") $
       failIO "Der neue Name mu ein reiner Dateiname, ohne Pfad, sein."
   when (zielpfad == symlinkpfad) $
       failIO "Symlink und Linkziel sind im gleichen Verzeichnis. Kann nicht beiden den gleichen Namen geben."

   mv linkziel neuziel
   rm symlink
   HsUnix.symlink neuziel neusymlink

pruefe_symlink symlink = do
    issym <- is_symlink symlink
    if issym then return ()
             else hPutStrLn stderr ("Kein Symlink, bersprungen: " ++ shell_quote symlink)
                  >> throwDyn Weiter

pruefe_linkziel symlink = do
    pruefe_symlink symlink
    r <- readlink symlink
    ex <- path_exists r
    if ex then return ()
          else hPutStrLn stderr ("Baumelnder Symlink, bersprungen: " ++ shell_quote symlink)
               >> throwDyn Weiter

-- Wenn der Symlink in einen Behlter zeigt, dann mu der offen sein. Ist
-- er semioffen, so wird er geffnet. Ist er geschlossen, so wird eine
-- Meldung ausgegeben und eine dynamische Weiter-Ausnahme geworfen.
pruefe_offen :: String  -- Symlink
             -> IO Bool -- Ob er in einen Behlter zeigt
pruefe_offen symlink = do
   linkziel <- readlink symlink
   let linkziel_abs       = absolute_path' linkziel (fst (split_path symlink))
       (behverz, cdverz0) = behverz_aus_abspfad linkziel_abs                    -- behverz: Behlterverzeichnis, in das der Symlink zeigt
                                                                                -- cdverz0: Behltername im Behlterverzeichnis
       cdverz             = behverz ++ "/" ++ cdverz0                           -- cdverz:  voller Pfad zum Behlter
   if behverz == ""
      then return False
      else do o <- beh_offen cdverz
              s <- beh_semioffen cdverz
              when (not o && s) $ do
                 beh_semioeffnen True False cdverz
              if (o || s)
                 then return True
                 else do hPutStrLn stderr $ "\nSymlink zeigt in einen geschlossenen Behlter. berspringe ihn.\n\
                                            \Symlink:  " ++ shell_quote symlink ++ "\n\
                                            \Linkziel: " ++ shell_quote linkziel_abs ++ "\n"
                         >> throwDyn Weiter


kzp = do
     args <- getargs "umlinken <Optionen> <Symlinks>" descs
     args_one [ d_ersetzen, d_herholen, d_tief_loeschen, d_tief_umbenennen ] args
     let fkt = if arg_switch args d_ersetzen then Ersetzen
                else if arg_switch args d_herholen then Herholen
                 else if arg_switch args d_tief_loeschen then Tief_loeschen
                  else if arg_switch args d_tief_umbenennen then Tief_umbenennen
                   else error "Bug in kzp in umlinken.hs"
         direct = args_req args d_direct
     return (fkt, direct)
